/*
** Module   : ChuiSimulator.java
** Abstract : Swing-based terminal simulator
**
** Copyright (c) 2011-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20110921 Created initial version.
** 002 GES 20111026 Added safety code to only dispose the frame when it is
**                  displayable.
** 003 EVL 20121119 Adding code to initially select the first available
**                  monospaced font when application starts.
** 004 MAG 20131114 Change this class to became a base for ChuiWebSimulator.
** 005 MAG 20131201 Fix javadoc errors.
** 006 MAG 20140114 Add protected methods used by ChuiWebSimulator.
*/

package com.goldencode.p2j.ui.client.chui.driver.swing;

import java.awt.*;
import java.awt.event.*;
import java.awt.font.*;
import java.awt.image.*;
import java.awt.print.*;
import javax.swing.*;
import java.util.*;

import com.goldencode.p2j.ui.client.swing.*;

// we import selectively because there are many overlapping class names between
// swing and p2j, since the current class is more swing than p2j we err on the
// side of swing imports
import com.goldencode.p2j.ui.client.Cell;

/**
 * Swing-based terminal simulator. This control is the core of a rectangular
 * component that appears to look and act like a character terminal (TTY).
 * <p>
 * All output is done in the same fixed width font family, with character
 * attributes implemented by modifying the attributes of the common font for
 * each output character.  The output stored in a 2-dimensional buffer of
 * of cells, which store the character to be drawn and the attributes/color
 * of that character. 
 * <p>
 * The current cursor position is remembered and can be manipulated by the
 * application. The cursor visibility state can be externally controlled as
 * well. When rendering output, if the cursor is visible, it is drawn by
 * inverting the foreground/background colors of the font used to draw the
 * character at which the cursor is located.   
 * <p>
 * Input is read using key events and that is kept in a FIFO buffer as if
 * they had been sent by the remote terminal hardware.
 * <p>
 * Other common end-user functionality is implemented including printing,
 * selection of text, copy of selected text to the clipboard, paste of text
 * from the clipboard into the keyboard input buffer and copying the entire
 * terminal's visible state to the clipboard as an image.
 * <p>
 * Limitations:
 * <p>
 * <ul>
 *    <li> The BLINK attribute is not supported yet. To implement it, one would
 *         have to keep track (in a carefully synchronized way) of the list of
 *         characters (and the current font being used) that must be blinked
 *         periodically. Then a timer would have to walk that list periodically
 *         and invert the colors (using TextAttribute.SWAP_COLORS).
 *    <li> Child processes which are interactive (they read and write from the
 *         terminal/pty) will NOT WORK. There is no easy solution for this
 *         issue. When run as an application, non-interactive child process
 *         launching and usage works without issue.
 *    <li> This class doesn't use Swing's double buffering yet. It may not need
 *         it ever. The redrawing seems pretty clean.
 * </ul>
 */
@SuppressWarnings("serial")
public class ChuiSimulator
extends JComponent
implements Printable
{
   /** Immutable empty cell (space character with normal color). */
   private final static Cell EMPTY_CELL =
      new Cell(' ', com.goldencode.p2j.ui.Color.NORMAL);
   
   /** Lock object for synchronizing configuration usage and changes. */
   private Object cfgLock = new Object();
   
   /** Lock object for synchronizing selection processing. */
   private Object selLock = new Object();
   
   /** Lock object for synchronizing data and cursor usage and changes. */
   private Object dataLock = new Object();
   
   /** Terminal configuration storage. */
   private TerminalOptions options = null;
   
   /** Helper to map color specifications to the font needed for drawing. */
   private ColorMapper colorMap = null;
   
   /** Popup menu. */
   private TerminalMenu menu = null;
   
   /** The height of a single character in pixels (same as the row height). */
   private int charHeight = 0;
   
   /** The width of a single character in pixels. */
   private int charWidth = 0;
   
   /** Number of pixels between each row based on font metrics. */
   private int baseLine = 0;
   
   /** Selection start X coordinate in pixels. */
   private int x0 = -1;
   
   /** Selection end X coordinate in pixels. */
   private int x1 = -1;
   
   /** Selection start Y coordinate in pixels. */
   private int y0 = -1;
   
   /** Selection end Y coordinate in pixels. */
   private int y1 = -1;
   
   /** Selection start row number. */
   private int r0 = -1;
   
   /** Selection end row number. */
   private int r1 = -1;
   
   /** Selection start column number. */
   private int c0 = -1;
   
   /** Selection end column number. */
   private int c1 = -1;
   
   /** Screen contents. */
   private Cell[][] data = null;
   
   /** Current X coordinate of the cursor. */
   private int cursorX = 0;

   /** Current Y coordinate of the cursor. */
   private int cursorY = 0;
   
   /** Flag that controls if the cursor is visible or not. */
   private boolean showCursor = true;
   
   /** Flag to control if it is valid to track selection events. */
   private boolean selectionTrack = false;
   
   /** Boundary of the current selected region in pixel coordinates. */ 
   private Rectangle selected = null;
   
   /** Key processing helper instance. */
   private SwingKeyboardReader keyReader = new SwingKeyboardReader();
   
   /** Containing frame (may be an applet or an application). */
   private Frame frame = null;
   
   /**
    * Default constructor.    
    */
   public ChuiSimulator()
   {      
   }
   
   /**
    * Create a new instance with the container and frame given. The component's
    * size and characteristics will be determined from the options provided.
    * When this method is done, the preferred size of the component will have
    * been established, all listeners and key processing will be initialized
    * and the control will be ready to be made visible by the caller.
    * <p>
    * For applications, the frame and the container are usually the same
    * object (of type <code>JFrame</code>). For applets, these are different
    * instances since a <code>JApplet</code> will be the container and the
    * frame will be a parent container of the <code>JApplet</code>.
    *
    * @param    frame
    *           The containing frame. Needed for key event processing
    *           and for the processing of modal dialog boxes.
    * @param    container
    *           The immediate container component, which can be called to
    *           force a resizing to match a changed simulator size.
    * @param    options
    *           The terminal configuration (size and appearance).
    */
   public ChuiSimulator(Frame           frame,
                        ContainerOps    container,
                        TerminalOptions options)
   {
      if(options.fontname.equalsIgnoreCase("monospaced"))
         options.fontname = SwingHelper.getMonoFontName(this, options.fontname);
      configure(options);
      
      this.frame = frame;
      
      // setup popup menus
      menu = new TerminalMenu(frame, this, container, options);
      addMouseListener((MouseListener) new PopupTrigger(menu));
      
      // enable selection tracking
      SelectionTracker tracker = new SelectionTracker(this);
      addMouseListener(tracker);
      addMouseMotionListener(tracker);
      
      // listen for keystrokes
      frame.addKeyListener(keyReader);
      
      // disable focus processing on TAB and SHIFT-TAB, otherwise those
      // keys are not ever available for consumption
      frame.setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
                                  new HashSet<AWTKeyStroke>());
      frame.setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS,
                                  new HashSet<AWTKeyStroke>());
   }
   
   /**
    * Capture and return the terminal screen as an image.
    *
    * @return   An image representing the terminal screen that is being
    *           displayed.
    */
   public Image capture()
   {
      BufferedImage image = new BufferedImage(getWidth(),
                                              getHeight(),
                                              BufferedImage.TYPE_4BYTE_ABGR);
      
      printAll(image.createGraphics());
      
      return image;
   }
   
   /**
    * Create a new printer job, open the job dialog for the user and if
    * printing is not canceled by the user, print the current terminal screen
    * to the chosen printer, with the configured job settings.
    */
   public void print()
   {
      PrinterJob job = PrinterJob.getPrinterJob();
      job.setPrintable(this);
      
      if (job.printDialog())
      {
         try
         {
            job.print();
         }
         
         catch (PrinterException pe)
         {
            // ignore
         }
      }      
   }
   
   /**
    * Low level printing worker that is called back by the Java printer job
    * in order to generate the print output. This will only ever print a single
    * page of output. It does not scale the print output to the page, so it
    * is possible that the output will be cut off to the right and on the
    * bottom.
    *
    * @param    g
    *           The graphics instance into which to draw.
    * @param    pf
    *           Details of the printer page being processed.
    * @param    page
    *           The page number being requested. Only the 0 page will ever
    *           be honored.
    *
    * @return   <code>PAGE_EXISTS</code> if the page was printed or
    *           <code>NO_SUCH_PAGE</code> if any page number other than 0
    *           is requested.
    */
   public int print(Graphics g, PageFormat pf, int page)
   throws PrinterException
   {
      if (page > 0)
      {
         return NO_SUCH_PAGE;
      }

      Graphics2D graphics = (Graphics2D) g;
      graphics.translate(pf.getImageableX(), pf.getImageableY());

      printAll(graphics);

      return PAGE_EXISTS;
   }   
   
   /**
    * Cause the outer frame to be disposed, exiting the application or applet.
    * This is important since <code>System.exit()</code> cannot be used for
    * applets.
    */
   public void quit()
   {
      // protect against multiple calls in the same session
      if (frame.isDisplayable())
         frame.dispose();
   }
   
   /**
    * Read the next key from the FIFO key input buffer.  This will block until
    * a key is available.
    *
    * @return   The next key in the buffer.
    */
   public int readKey()
   {
      return keyReader.read();
   }
   
   /**
    * Copy the given cells into the 2-dimensional output buffer starting at
    * the current cursor position and writing to the "right", incrementing
    * the cursor X position for each cell. When the last column is reached,
    * the cursor X will wrap back to 0 and the cursor Y will increment to the
    * next row. When the last column of the last row is written, both the
    * cursor X and Y positions will wrap back to 0.
    * <p>
    * This write operation does not cause a repaint to occur, but the next
    * repaint that does occur will do so with the updated content.
    *
    * @param    output
    *           List of cells to write into the buffer.
    */
   public void append(Cell[] output)
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         w = options.columns;
         h = options.rows;
      }
      
      synchronized (dataLock)
      {
         for (Cell next : output)
         {
            // copy the data to the current cursor position and increment the
            // X cursor
            data[cursorY][cursorX++] = next;
            
            // wrap our cursor if needed
            if (cursorX == w)
            {
               cursorX = 0;
               cursorY++;
               
               if (cursorY == h)
               {
                  cursorY = 0;
               }
            }
         }
      }
   }
   
   /**
    * Completely replace the current 2-dimensional output buffer with the
    * content on the given <code>screen</code> parameter and then set the
    * the cursor position to the given values. Only non-null cells are copied.
    * Any cells that are null are ignored and left alone in the target buffer.
    * <p>
    * This write operation DOES cause a repaint to occur.
    *
    * @param    screen
    *           2-dimensional array of cells to write into the buffer. This
    *           must have the same dimensions as the output buffer and all
    *           rows must have column arrays of the same size.
    * @param    newX
    *           Cursor X position after the method call.
    * @param    newY
    *           Cursor Y position after the method call.
    */
   public void replace(Cell[][] screen, int newX, int newY)
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         w = options.columns;
         h = options.rows;
      }
      
      synchronized (dataLock)
      {
         // safety code
         if (screen.length == h && data.length == h)
         {
            for (int row = 0; row < h; row++)
            {
               // more safety code
               if (screen[row].length == w && data[row].length == w)
               {
                  for (int col = 0; col < w; col++)
                  {
                     // no need to clone each cell since the cell class' data
                     // is immutable
                     if (screen[row][col] != null)
                     {
                        data[row][col] = screen[row][col];
                     }
                  }
               }
            }
         }
         
         setCursorWorker(newX, newY, w, h);
      }
      
      triggerRepaint(true);
   }
   
   /**
    * Overwrite the entire 2-dimensional output buffer with the space character
    * to be rendered with the <code>Color.NORMAL</code> color. Sets the cursor
    * Y and X positions to 0.
    */
   public void clear()
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         w = options.columns;
         h = options.rows;
      }
      
      synchronized (dataLock)
      {
         for (int row = 0; row < h; row++)
         {
            for (int col = 0; col < w; col++)
            {
               data[row][col] = EMPTY_CELL;
            }
         }
         
         setCursorWorker(0, 0, w, h);
      }
   }
   
   /**
    * Position the cursor.
    *
    * @param    x
    *           New X cursor position.
    * @param    y
    *           New Y cursor position.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   public boolean setCursor(int x, int y)
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         w = options.columns;
         h = options.rows;
      }
      
      return setCursorWorker(x, y, w, h);
   }
   
   /**
    * Set the cursor visibility.
    *
    * @param    on
    *           <code>true</code> to make the cursor visible (subsequen repaint
    *           operations will then cause the cursor to be drawn).
    *
    * @return   The previous cursor visiblity state.
    */
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException
   {
      boolean old = false;
      
      synchronized (dataLock)
      {
         old = showCursor;
         showCursor = on;
      }
      
      return old;
   }
   
   /**
    * Get the cursor's current X coordinate.
    *
    * @return   The current X coordinate.
    */
   public int getCursorX()
   {
      int x = -1;
      
      synchronized (dataLock)
      {
         x = cursorX;
      }
      
      return x;
   }
   
   /**
    * Get the cursor's current Y coordinate.
    *
    * @return   The current Y coordinate.
    */
   public int getCursorY()
   {
      int y = -1;
      
      synchronized (dataLock)
      {
         y = cursorY;
      }
      
      return y;
   }
   
   /**                                         
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   public int getRows()
   {
      int r = -1;
      
      synchronized (cfgLock)
      {
         r = options.rows;
      }
      
      return r;
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   public int getColumns()
   {
      int c = -1;
      
      synchronized (cfgLock)
      {
         c = options.columns;
      }
      
      return c;
   }

   /**
    * Initialize the instance using the given configuration. This will obtain
    * and initialize its fonts (and the color mapper). It will use the font
    * metrics to calculate the size of the component (based on the number of
    * rows and columns).  If this is the first initialization, the 2-dimensional
    * output buffer will be created and cleared.
    *
    * @param    terminalOptions
    *           The configuration for the terminal simulator.
    */
   public void configure(TerminalOptions terminalOptions)                    
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         this.options = (terminalOptions == null) ? new TerminalOptions() : terminalOptions;
         
         // this part doesn't seem to really work
         setOpaque(true);
         setBackground(options.background);
         setForeground(options.foreground);
         
         Map<TextAttribute, Object> fontAttrs = new HashMap<TextAttribute, Object>();
         fontAttrs.put(TextAttribute.FAMILY, options.fontname);
         fontAttrs.put(TextAttribute.SIZE, options.fontsize);
         fontAttrs.put(TextAttribute.FOREGROUND, options.foreground);
         fontAttrs.put(TextAttribute.BACKGROUND, options.background);
                                                                     
         Font font = new Font(fontAttrs);
         
         if (colorMap == null)
         {
            colorMap = new ColorMapper(font);
         }
         else
         {
            colorMap.reset(font);
         }
         
         // save off the "official" font family name
         options.fontname = font.getFamily();
         
         setFont(font);
         
         FontMetrics fm = getFontMetrics(font);
         
         // must reduce this value by 1 otherwise there is a 1 pixel blank line
         // between rows
         charHeight = fm.getHeight() - 1;
         
         // previous attempts to use fm.getMaxAdvance() - 1 did not work for
         // all fonts, this hack seems to work well instead
         charWidth  = fm.charWidth('m');
         
         baseLine   = charHeight - fm.getMaxAscent();
         
         if (!SwingHelper.isMonospaced(font, fm))
         {
            throw new IllegalArgumentException("Font face '" + 
                                               options.fontname +
                                               "' must be fixed width.");
         }
         
         w = options.columns;
         h = options.rows;
         
         Dimension sz = new Dimension((charWidth * options.columns),
                                      (charHeight * options.rows));
         setPreferredSize(sz);
         setMinimumSize(sz);
         setMaximumSize(sz);
      }
      
      synchronized (dataLock)
      {
         if (data == null)
         {
            data = new Cell[h][w];
            clear();
         }
      }
   }
   
   /**
    * Paint the entire component to the given graphics instance, using
    * previously set state.  Depending on the repaint request, the graphics
    * object's clipping region may have been set to a subset of the full
    * rectangle. This method makes no such distinction, it just blindly
    * draws everything and the stuff outside of the clipping region never
    * gets sent to the hardware.  This is the core where all drawing occurs
    * and it draws the screen text (including switching out the fonts per
    * character as needed to get the right colors/attributes), the current
    * cursor (if it is visible) and the any selection box (when a selection
    * is active).
    *
    * @param    graphics
    *           The object into which to draw.
    */
   public void paintComponent(Graphics graphics)
   {
      super.paintComponent(graphics);
      
      int         rows = -1;
      int         cols = -1;
      int         base = -1;
      int         h    = -1;
      int         w    = -1;
      ColorMapper cm   = null;
      
      synchronized (cfgLock)
      {
         rows = options.rows;
         cols = options.columns;
         base = baseLine;
         h    = charHeight;
         w    = charWidth;
         cm   = colorMap;
      }
      
      // save off font
      Font oldFont = graphics.getFont();
      
      synchronized (dataLock)
      {
         // draw each row
         for (int i = 0; i < rows; i++)
         {
            // calculate the Y position (in pixels) for this row
            int y = ((i + 1) * h) - base - 1;
            
            int j = 0;
            int x = 0;
            
            Font last = null;
            Font font = null;
            StringBuilder sb = new StringBuilder();
            
            while (j < cols)
            {
               Cell next = data[i][j];
               char ch   = next.character();
               
               font = cm.toFont(next.color());
               
               if (last != null && last != font)
               {
                  // flush our current text
                  drawString(graphics, sb.toString(), last, x, y);
                  
                  // clear our text buffer
                  sb = new StringBuilder();
                  
                  // recalculate x
                  x = j * w;
               }
               
               // accumulate the text since it is either the first in a
               // new buffer or it shares the same font as the previous text
               sb.append(Character.isISOControl(ch) ? ' ' : ch);
               
               // save off the most recent font
               last = font;
               
               // look at the next char
               j++;
            }
            
            if (sb.length() > 0)
            {
               // output the remaining text in the buffer
               drawString(graphics, sb.toString(), font, x, y);
            }
         }
      }
      
      // draw cursor (if needed)
      if (showCursor)
      {
         StringBuilder cursor = new StringBuilder();
         cursor.append(data[cursorY][cursorX].character());
         int y = ((cursorY + 1) * h) - base - 1;
         int x = cursorX * w;
         
         Font fnt = cm.toFontReverse(data[cursorY][cursorX].color());
         drawString(graphics, cursor.toString(), fnt, x, y);
      }
      
      // restore font
      graphics.setFont(oldFont);
      
      Rectangle area = null;
      
      synchronized (selLock)
      {
         area = selected;
      }
      
      if (area != null)
      {
         Color old = graphics.getColor();
         graphics.setColor(options.selection);
         graphics.drawRect((int) area.getX(),
                           (int) area.getY(),
                           (int) area.getWidth(),
                           (int) area.getHeight());
         graphics.setColor(old);
      }
   }
   
   /**
    * Read the current terminal buffer at each character position inside the
    * current selection and return the text in a string. Each row of the
    * selection (except for the last one) will be terminated with an embedded
    * newline character inside the resulting string.
    *
    * @return   The lines of text on the screen inside the current selection
    *           or an empty string if there is no current selection.
    */
   public String copy()
   {
      StringBuilder sb = new StringBuilder();
      
      int r0 = -1;
      int r1 = -1;
      int c0 = -1;
      int c1 = -1;
      
      synchronized (selLock)
      {
         r0 = this.r0;
         r1 = this.r1;
         c0 = this.c0;
         c1 = this.c1;
      }
      
      synchronized (dataLock)
      {
         for (int i = r0; i <= r1; i++)
         {
            for (int j = c0; j <= c1; j++)
            {
               sb.append(data[i][j].character());
            }
            
            if (i != r1)
               sb.append('\n');
         }
      }
      
      // now clear the current selection (this will recursively call back into
      // the simulator to trigger a cancelSelection() which will clear state)  
      menu.textSelectionDone();
      
      return sb.toString();
   }
   
   /**
    * Insert the given characters into the keyboard input buffer (at the end,
    * in FIFO order), then call {@link #cancelSelection}.
    *
    * @param    data
    *           Characters to insert into the "keyboard" buffer.
    */
   public void paste(String data)
   {
      keyReader.insert(data);
      
      // clear the current selection
      cancelSelection();
   }
   
   /**
    * Control whether the selection tracker is tracking mouse events.
    *
    * @param    track
    *           <code>true</code> to allow tracking, <code>false</code> to
    *           ignore all mouse events (no tracking).
    */
   public void enableSelectionTrack(boolean track)
   {
      synchronized (selLock)
      {
         selectionTrack = track;
      }
   }
   
   public boolean canSelectionTrack()
   {
      boolean track = false;
      
      synchronized (selLock)
      {
         track = selectionTrack;
      }
      
      return track;
   }
   
   /**
    * Start selection processing based on the given starting coordinates and
    * force a repaint of the selection region which will cause the selection
    * box to be drawn around the character in which the point is found. This
    * will also cause dependent popup menu items to be enabled.
    *
    * @param    x
    *           Most recent ending X coordinate.
    * @param    y
    *           Most recent ending Y coordinate.
    */
   public void startSelection(int x, int y)
   {
      synchronized (selLock)
      {
         x0 = x1 = x;
         y0 = y1 = y;
         selected = calcCharBoundary();
         triggerRepaint(true);
      }
      menu.enableCopy(true);
   }
   
   /**
    * Update the selection state based on the latest ending coordinates and
    * force a repaint of the selection region which will cause the selection
    * box to be drawn in its new position and size.
    *
    * @param    x
    *           Most recent ending X coordinate.
    * @param    y
    *           Most recent ending Y coordinate.
    */
   public void updateSelection(int x, int y)
   {
      synchronized (selLock)
      {
         // must be checked before edits are made
         boolean full = isInsideSelectionArea(x, y);
         
         x1 = x;
         y1 = y;
         selected = calcCharBoundary();
         triggerRepaint(full);
      }
   }
   
   /**
    * End the current selection operation, update the final selection state and
    * force a repaint of the selection region which will cause the selection
    * box to be drawn in its final position and size.
    *
    * @param    x
    *           Ending X coordinate.
    * @param    y
    *           Ending Y coordinate.
    */
   public void endSelection(int x, int y)
   {
      synchronized (selLock)
      {
         x1 = x;
         y1 = y;
         selected = calcCharBoundary();
         triggerRepaint(false);
      }
   }
   
   /**
    * Cancel the current selection operation, reset the selection state and
    * force a repaint of the entire screen which will cause the selection
    * box to disappear. This will also cause the dependent popup menu items
    * to be disabled.
    */
   public void cancelSelection()
   {
      synchronized (selLock)
      {
         if (selected != null)
         {
            x0 = x1 = -1;
            y0 = y1 = -1;
            selected       = null;
            selectionTrack = false;
            triggerRepaint(true);
         }
      }
      
      menu.enableCopy(false);
   }
   
   /**
    * Cause a partial or full repaint to occur. If <code>full</code> is
    * <code>true</code> or if there is no current selection operation occurring,
    * a repaint of the entire component ("screen") will occur.  Otherwise,
    * a repaint of only the selection area will occur.
    *
    * @param    full
    *           Flag to force a the entire screen to repaint.
    */
   public void triggerRepaint(boolean full)
   {
      // calculate the selection box as a rectangle
      Rectangle area = null;
      
      if (selected == null || full)
      {
         area = new Rectangle(getPreferredSize());
      }
      else
      {
         // we must redraw one more pixel in height and width than the actual
         // rectangle, since the drawn rectangle includes 1 extra pixel to the
         // right and below in each dimension
         area = new Rectangle((int) selected.getX(),
                              (int) selected.getY(),
                              (int) selected.getWidth() + 1,
                              (int) selected.getHeight() + 1);
      }
      
      // trigger the painting
      repaint(area);
   }
   
   /**
    * Helper method to draw the given string into the given graphics object
    * at the given location with the given font.
    *
    * @param    graphics
    *           Instance to draw into.
    * @param    text
    *           Text to draw.
    * @param    font
    *           The common font in which to draw all characters from the string.
    * @param    x
    *           X coordinate.
    * @param    y
    *           Y coordinate.
    */
   private void drawString(Graphics graphics, String text, Font font, int x, int y)
   {
      graphics.setFont(font);
      graphics.drawString(text, x, y);      
   }
   
   /**
    * Detect if the given point is inside the current selection rectangle.
    *
    * @param    x
    *           X coordinate to check.
    * @param    y
    *           Y coordinate to check.
    *
    * @return   <code>true</code> if the point is inside the selection area.
    */
   private boolean isInsideSelectionArea(int x, int y)
   {
      return (x0 < x1 && x0 < x && x < x1) ||
             (x0 > x1 && x0 > x && x > x1) ||
             (y0 < y1 && y0 < y && y < y1) ||
             (y0 < y1 && y0 < y && y < y1);
   }
   
   /**
    * Based on the X and Y coordinates in pixels from a mouse button event
    * (during selection processing), calculate the starting and ending
    * coordinates of the selection rectangle in both character positions
    * (rows/columns) and in pixels. The pixel positions will "snap" to
    * locations that are on the upper left corner of the top leftmost selected
    * character and to the lower right corner of the bottom rightmost selected
    * character. The calculated character boundaries are cached for future
    * use.
    *
    * @return   The rectangle that defines the pixel boundaries of this
    *           area.
    */
   private Rectangle calcCharBoundary()
   {                                                           
      Rectangle area = null;                                   
      
      int tx0 = this.x0;
      int ty0 = this.y0;
      int tx1 = this.x1;
      int ty1 = this.y1;
      
      if (tx0 != -1)
      {
         // normalize the values so that (x0, y0) is the upper left and
         // (x1, y1) is the lower right
         int x0 = Math.min(tx0, tx1);
         int x1 = Math.max(tx0, tx1);
         int y0 = Math.min(ty0, ty1);
         int y1 = Math.max(ty0, ty1);
         
         // calculate in which character (row and column) each point exists
         // (note that the use of integer division will naturally leave us at
         // the character boundries)
         r0 = y0 / charHeight;
         c0 = x0 / charWidth;
         r1 = y1 / charHeight;
         c1 = x1 / charWidth;
         
         int x = c0 * charWidth;
         int y = r0 * charHeight;
         int w = ((c1 + 1) - c0) * charWidth;
         int h = ((r1 + 1) - r0) * charHeight;
         
         area = new Rectangle(x, y, w, h);
      }
      
      return area;
   }
   
   /**
    * Position the cursor.
    *
    * @param    x
    *           New X cursor position.
    * @param    y
    *           New Y cursor position.
    * @param    w
    *           Screen width.
    * @param    h
    *           Screen height.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   private boolean setCursorWorker(int x, int y, int w, int h)
   {
      if (x < 0 || x >= w || y < 0 || y >= h)
         return false;
      
      synchronized (dataLock)
      {
         cursorX = x;
         cursorY = y;
      }
      
      return true;
   }
   
   /**
    * Get screen memory array.
    * 
    * @return  The screen memory array.
    */
   protected Cell[][] getScreenData()
   {
      return data;
   }

   /**
    * Get terminal options.
    * 
    * @return  The terminal options.
    */
   protected TerminalOptions getTerminalOptions()
   {
      return options;
   }
}
