/*
** Module   : ChuiWebDriver.java
** Abstract : WebSocket Console-based driver.
**
** Copyright (c) 2013-2014 Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 MAG 20131112 Created initial version.
** 002 MAG 20131129 Import server KeyStore.
** 003 MAG 20140109 Add Dojo modules handler and resource handler.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

import java.util.logging.*;

import org.eclipse.jetty.server.handler.*;
import org.eclipse.jetty.util.resource.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.web.*;

/**
 * Chui WebSocket console-based driver.
 *
 */
public class ChuiWebDriver
extends ChuiScreenDriver
implements WebScreenDriver
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ChuiWebDriver.class.getName());

   /** Embedded web server */
   private GenericWebServer server = new GenericWebServer();

   /** Main page root target. */
   private static final String MAIN_TARGET = "/";

   /** WebSocket target. */
   private static final String AJAX_TARGET = "/ajax";
   
   /** Configuration. */
   private BootstrapConfig config;
   
   /** chui simulator. */
   private ChuiWebSimulator simulator;
   
   /**
    * Constructor.
    * 
    * @param   config
    *          The <code>BootstrapConfig</code> configuration.
    */
   public ChuiWebDriver(BootstrapConfig config)
   {
      this.config = config;
   }
   
   /**
    * Initialize driver.
    */
   @Override
   public void init()
   {
      // get terminal options
      simulator = new ChuiWebSimulator();
      TerminalOptions options = (config == null) ? 
                                 new TerminalOptions() : 
                                 ChuiScreenDriver.convertToOptions(config);
      simulator.configure(options);
      setChuiSimulator(simulator);
      setPrimitives(new ChuiPrimitives(simulator));      
   }

   /**
    * Startup the embedded server using server KeyStore to build 
    * the server SSL context.
    * 
    * @param   keyStore
    *          The exported server KeyStore. 
    */
   public String startupServer(ServerKeyStore keyStore)
   {
      String uri = null;
      // startup embedded server
      try
      {
         // welcome page
         String welcomePageUrl = getWelcomePageUrl("/", "/index.html");
         ChuiWebPageHandler mainPage = new ChuiWebPageHandler(config, 
                                                              MAIN_TARGET, 
                                                              welcomePageUrl);
         // configure and startup server
         server.setKeyStore(keyStore);
         server.setBootstrapConfig(config);
         server.addHandler(mainPage);
         server.addHandler(new DojoToolkitHandler());
         
         ResourceHandler resHandler = new ResourceHandler();         
         StringBuilder sb = new StringBuilder("/");
         String packageName = WebScreenDriver.class.getPackage().getName();
         sb.append(packageName.replace('.', '/')).append("/res").toString();
         resHandler.setBaseResource(Resource.newClassPathResource(sb.toString()));
         server.addHandler(resHandler);
         
         server.addWebSocketHandler(AJAX_TARGET, 
                  new GenericWebSocketCreator<ChuiSimulator>(simulator));
         
         server.startup();
      }
      catch (Exception e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebDriver.init()",
                  "",
                  LogHelper.generate("Cannot startup server."),
                  e);         
      }
      finally
      {
         uri = this.getServerURI();
      }
      
      return uri;
   }
   
   /**
    * Build the name of the main page as a resource locator string.
    * 
    * @param   pageContext
    *          Page context root.
    * @param   pageFile
    *          Main page HTML file.
    * 
    * @return  Resource name for the main page.
    */
   private String getWelcomePageUrl(String pageContext, String pageFile)
   {
      // build main page resource name
      StringBuilder sb = new StringBuilder(pageContext);
      String packageName = ChuiWebPageHandler.class.getPackage().getName();
      return sb.append(packageName.replace('.', '/')).append(pageFile).toString();
   }   

   /**
    * Get server URI
    * 
    * @return  A string representing the server URI.
    */
   public String getServerURI()
   {
      String uri = null;
      
      if (server.isRunning())
      {
         uri = String.format("https://%s:%d%s", 
                             server.getHost(), 
                             server.getPort(), 
                             server.getContextPath());
      }
      
      return uri;
   }   
   
   /**
    * Emit a sound beep on the terminal.
    * 
    */
   @Override
   public void beep()
   {
      simulator.beep();
   }
}
